# Copyright 2023 Inductor, Inc.
"""Functionality for test suite execution tasks."""

from concurrent import futures
import contextlib
import datetime
from typing import Dict, Generator, Iterator, List, Literal, NamedTuple, Optional, Union

import inquirer
import rich
from rich import progress
import typer

from inductor import config
from inductor.backend_client import backend_client, wire_model
from inductor.data_model import data_model
from inductor.execution import execution, quality_measure_execution


@contextlib.contextmanager
def _disable_logger_decorator():
    """Disable the inductor.logger decorator within this context manager.

    Disable the inductor.logger decorator by setting
    `execution.logger_decorator_enabled` to False. On exit, restore the
    original value of `execution.logger_decorator_enabled`.
    """
    orig_logger_decorator_enabled = execution.logger_decorator_enabled
    try:
        execution.logger_decorator_enabled = False
        yield
    finally:
        execution.logger_decorator_enabled = orig_logger_decorator_enabled


class _ExecuteTestCaseReturn(NamedTuple):
    """Return type of `execute_test_case`.

    See `execute_test_case`'s docstring (below) for
    descriptions of fields.
    """
    execution_details: wire_model.ExecutionDetails
    invalid_quality_measures: List[Dict[str, Union[
        int, str, wire_model.QualityMeasureExecutionDetails]]]
    stdout_stderr: str


def execute_test_case(
    llm_program_fully_qualified_name: str,
    test_case: data_model.TestCase,
    quality_measures_with_metadata: Optional[List[
        data_model.QualityMeasureWithMetadata]] = None,
    hparams: Optional[Dict[str, Union[str, int, float]]] = None,
    mode: Literal["CLI", "PLAYGROUND"] = "CLI",
    suppress_stdout_stderr: bool = True
) -> _ExecuteTestCaseReturn:
    """Run a test case and evaluate its quality measures.

    Args:
        llm_program_fully_qualified_name: Fully qualified name of LLM program.
        test_case: Test case.
        quality_measures_with_metadata: List of quality measures with ID.
        hparams: Mapping from hyperparameter names to values.
        mode: Execution mode.
        suppress_stdout_stderr: Whether to suppress stdout and stderr output.
    
    Returns:
        Named tuple with the following fields:
            execution_details: ExecutionDetails object.
            invalid_quality_measures: List of invalid quality measures. Each
                invalid quality measure is a Dictionary with the following keys:
                    "id": ID of quality measure.
                    "name": Name of quality measure.
                    "execution_details": QualityMeasureExecutionDetails object.
            stdout_stderr: Combined stdout and stderr output, if
                suppress_stdout_stderr is False. Otherwise, an empty string.
    """
    started_at = datetime.datetime.now(datetime.timezone.utc)

    llm_program_stdout = None
    llm_program_stderr = None
    stdout_stderr = ""
    llm_program_error = None
    llm_program_output = None
    # Note: we use backslashes to split the `with` clause immediately below,
    # rather than enclosing in parentheses, because enclosing in parentheses
    # apparently causes a SyntaxError when running in Python 3.8.
    # pylint: disable-next=protected-access
    with execution.capture_stdout_stderr(
        suppress=suppress_stdout_stderr) as captured_streams, \
        execution.capture_logged_values() as logged_values, \
        _disable_logger_decorator(), \
        execution.set_hparams(hparams or {}):
        try:
            # Run the LLM program.
            llm_program_output = data_model.LazyCallable(
                llm_program_fully_qualified_name
            )(**test_case.inputs)
            if isinstance(llm_program_output, Iterator):
                llm_program_output = list(llm_program_output)
                if all(isinstance(value, str) for value in llm_program_output):
                    llm_program_output = "".join(llm_program_output)

        except Exception as error:  # pylint: disable=broad-except
            llm_program_error = str(error)

        llm_program_stdout = captured_streams.stdout.getvalue()
        llm_program_stderr = captured_streams.stderr.getvalue()
        stdout_stderr = captured_streams.suppressed.getvalue()

    ended_at = datetime.datetime.now(datetime.timezone.utc)

    # If the LLM program completed without error, run the executable quality
    # measures.
    if llm_program_error is None and quality_measures_with_metadata:
        quality_measure_execution_result = (
            quality_measure_execution.execute_quality_measures(
                test_case=test_case,
                llm_program_output=llm_program_output,
                quality_measures_with_metadata=quality_measures_with_metadata
            ))
        direct_evaluations = (
            quality_measure_execution_result.direct_evaluations)
        invalid_quality_measures = (
            quality_measure_execution_result.invalid_quality_measures)
        stdout_stderr += (
            quality_measure_execution_result.stdout_stderr)
    else:
        direct_evaluations = []
        invalid_quality_measures = []

    execution_details = wire_model.ExecutionDetails(
        mode=mode,
        inputs=test_case.inputs,
        hparams=hparams or None,
        output=llm_program_output,
        error=llm_program_error,
        stdout=llm_program_stdout,
        stderr=llm_program_stderr,
        execution_time_secs=(ended_at - started_at).total_seconds(),
        started_at=started_at,
        ended_at=ended_at,
        logged_values=logged_values or None,
        direct_evaluations=direct_evaluations or None)

    return _ExecuteTestCaseReturn(
        execution_details=execution_details,
        invalid_quality_measures=invalid_quality_measures,
        stdout_stderr=stdout_stderr)


@contextlib.contextmanager
def _manage_run_requests(
    test_suite: data_model.TestSuite,
    auth_access_token: str
) -> Generator[wire_model.CreateTestSuiteRunResponse, None, None]:
    """Send requests to the server to manage the creation/completion of a run.
    
    Send a request to the server to create a test suite run. Then, yield the
    response, which contains test suite run metadata. On exit, send a request
    to the server to mark the test suite run as complete.

    Args:
        test_suite: Test suite.
        auth_access_token: Auth0 access token.

    Yields:
        CreateTestSuiteRunResponse object.
    """
    test_suite_run = test_suite._get_run_request()  # pylint: disable=protected-access
    test_suite_run_metadata = backend_client.create_test_suite_run(
        test_suite_run, auth_access_token)
    try:
        yield test_suite_run_metadata
    finally:
        backend_client.complete_test_suite_run(
            wire_model.CompleteTestSuiteRunRequest(
                test_suite_run_id=
                    test_suite_run_metadata.test_suite_run_id,
                ended_at=
                    datetime.datetime.now(datetime.timezone.utc)),
            auth_access_token)


class _ExecuteTestSuiteTestCaseReturn(NamedTuple):
    """Return type of `_execute_test_suite_test_case`.

    See `_execute_test_suite_test_case`'s docstring (below)
    for descriptions of fields.
    """
    log_execution_request: wire_model.LogTestCaseExecutionRequest
    invalid_quality_measures: List[Dict[str, Union[
        int, str, wire_model.QualityMeasureExecutionDetails]]]
    stdout_stderr: str


def _execute_test_suite_test_case(
    test_suite_run_id: int,
    test_case_with_metadata: data_model.TestCaseWithMetadata,
    test_case_replica_index: int,
    hparams: Dict[str, Union[str, int, float]],
    llm_program_fully_qualified_name: str,
    quality_measures_with_metadata: List[
        data_model.QualityMeasureWithMetadata],
    auth_access_token: str,
) -> _ExecuteTestSuiteTestCaseReturn:
    """Execute a test suite test case and log the results to the server.
    
    Args:
        test_suite_run_id: ID of test suite run.
        test_case_with_metadata: Test case with metadata.
        test_case_replica_index: Index of test case replica.
        hparams: Mapping from hyperparameter names to values.
        llm_program_fully_qualified_name: Fully qualified name of LLM program.
        quality_measures_with_metadata: List of quality measures with ID.
        auth_access_token: Auth0 access token.
    
    Returns:
        Named tuple with the following fields:
            log_execution_request: LogTestCaseExecutionRequest object.
            invalid_quality_measures: List of invalid quality measures. Each
                invalid quality measure is a Dictionary with the following keys:
                    "id": ID of quality measure.
                    "name": Name of quality measure.
                    "execution_details": QualityMeasureExecutionDetails object.
            stdout_stderr: Combined stdout and stderr output.
    """
    test_case_results = execute_test_case(
        llm_program_fully_qualified_name=llm_program_fully_qualified_name,
        test_case=test_case_with_metadata,
        quality_measures_with_metadata=quality_measures_with_metadata,
        hparams=hparams)

    log_execution_request = wire_model.LogTestCaseExecutionRequest(
        test_suite_run_id=test_suite_run_id,
        test_case_id=test_case_with_metadata.id,
        test_case_replica_index=test_case_replica_index,
        execution_details=test_case_results.execution_details)

    backend_client.log_test_case_execution(
        log_execution_request, auth_access_token)

    return _ExecuteTestSuiteTestCaseReturn(
        log_execution_request=log_execution_request,
        invalid_quality_measures=test_case_results.invalid_quality_measures,
        stdout_stderr=test_case_results.stdout_stderr)


def execute_test_suite(
    test_suite: data_model.TestSuite,
    auth_access_token: str,
    *,
    prompt_open_results: bool = False
) -> wire_model.CreateTestSuiteRunResponse:
    """Execute a test suite.
    
    Execute a test suite while displaying relevant information to the user,
    including a progress bar.
    
    Args:
        test_suite: Test suite to execute.
        auth_access_token: Auth0 access token.
        prompt_open_results: Whether to prompt the user to open the test
            suite run results in a browser.
    
    Returns:
        CreateTestSuiteRunResponse object, containing test suite run metadata.
    """
    # Note: we use backslashes to split the `with` clause immediately below,
    # rather than enclosing in parentheses, because enclosing in parentheses
    # apparently causes a SyntaxError when running in Python 3.8.
    with futures.ProcessPoolExecutor(
            max_workers=test_suite.config.parallelize) as executor, \
        _manage_run_requests(
            test_suite, auth_access_token) as test_suite_run_metadata:

        # Note that `test_suite_run_metadata.test_case_ids` is the IDs
        # of `test_suite.test_cases` followed by the IDs of
        # `test_suite_run_metadata.imported_test_cases`. This ordering
        # needs to be preserved when updating the components with IDs.
        test_cases_with_metadata = data_model.components_with_ids(
            test_suite.test_cases +
            (test_suite_run_metadata.imported_test_cases or []),
            test_suite_run_metadata.test_case_ids,
            data_model.TestCaseWithMetadata)
        quality_measures_with_metadata = data_model.components_with_ids(
            test_suite.quality_measures,
            test_suite_run_metadata.quality_measure_ids,
            data_model.QualityMeasureWithMetadata)

        def pluralize(count: int) -> str:
            return "s" if count != 1 else ""
        rich.print(f"Running {len(test_cases_with_metadata)} test case"
              f"{pluralize(len(test_cases_with_metadata))} "
              f"({test_suite.config.replicas} run"
              f"{pluralize(test_suite.config.replicas)} each) ...")

        test_case_futures = []
        for test_case_replica_index in range(test_suite.config.replicas):
            for hparams in test_suite.hparam_combinations:
                for test_case_with_metadata in test_cases_with_metadata:
                    test_case_futures.append(executor.submit(
                        _execute_test_suite_test_case,
                        test_suite_run_id=
                            test_suite_run_metadata.test_suite_run_id,
                        test_case_with_metadata=test_case_with_metadata,
                        test_case_replica_index=test_case_replica_index,
                        hparams=hparams,
                        llm_program_fully_qualified_name=
                            test_suite.config.llm_program,
                        quality_measures_with_metadata=
                            quality_measures_with_metadata,
                        auth_access_token=auth_access_token
                    ))

        typer.echo(f"Go to {test_suite_run_metadata.url} to view results.")
        if prompt_open_results:
            open_url = inquirer.confirm(
                message="Open in browser?", default=True)
            if open_url:
                typer.launch(test_suite_run_metadata.url)

        # Display progress bar and optionally print test outputs.
        with progress.Progress() as progress_bar:
            progress_task = progress_bar.add_task(
                "Test Cases",
                total=(test_suite.config.replicas *
                        len(test_cases_with_metadata) *
                        len(test_suite.hparam_combinations)))

            for future in futures.as_completed(test_case_futures):
                progress_bar.advance(progress_task)
                test_case_result = future.result()
                test_output = test_case_result.log_execution_request
                invalid_quality_measures = (
                    test_case_result.invalid_quality_measures)
                stdout_stderr = test_case_result.stdout_stderr

                if stdout_stderr:
                    progress_bar.console.print(stdout_stderr)
                if test_output.execution_details.error is not None:
                    progress_bar.console.print(
                        "\n[red][bold][ERROR] Test case execution "
                        "raised an exception.[/bold] The following "
                        "execution will be recorded as FAILED and "
                        "quality measures will not be evaluated:[/red]")
                    progress_bar.console.print(test_output)
                elif config.verbose:
                    progress_bar.console.print(test_output)
                if invalid_quality_measures:
                    progress_bar.console.print(
                        "\n[red][bold][ERROR] One or more quality "
                        "measures raised an exception or returned an "
                        "invalid value.[/bold] The following quality "
                        "measures will not be recorded as part of this "
                        "test case execution:[/red]")
                    progress_bar.console.print(invalid_quality_measures)

        rich.print("Run complete.")
        return test_suite_run_metadata



# Copyright 2023 Inductor, Inc.
"""Inductor client library."""

from inductor import quality
from inductor.data_model.data_model import HparamSpec, QualityMeasure, TestCase, TestSuite, TestSuiteValueError
from inductor.execution.execution import hparam, log
from inductor.execution.live_execution import logger


__all__ = [
    "HparamSpec",
    "QualityMeasure",
    "TestCase",
    "TestSuite",
    "TestSuiteValueError",
    "hparam",
    "log",
    "logger",
    "quality"
]



# Copyright 2023 Inductor, Inc.
"""Functionality for executing quality measures."""

import traceback
from typing import Any, Dict, List, NamedTuple, Union

import openai
import tiktoken

from inductor.backend_client import wire_model
from inductor.data_model import data_model
from inductor.execution import execution


class Inputs(dict):
    """A dictionary that allows access to its keys as attributes."""

    def __getattribute__(self, key: str) -> Any:
        """Return the value corresponding to key in this dict.
        
        Args:
            key: The key to access.

        Raises:
            AttributeError: If the key does not exist in this dict.
        """
        try:
            return self[key]
        except KeyError as exc:
            raise AttributeError(key) from exc


def _openai_llm_quality_measure(
    output: str,
    inputs: Inputs,
    test_case: wire_model.TestCase,
    quality_measure: wire_model.QualityMeasure,
    model: str,
    prompt: str
) -> str:
    """Evaluate an LLM-powered quality measure using the OpenAI API.
    
    Args:
        output: The output of the execution of a LLM program.
        inputs: The inputs to the LLM program execution that produced output.
        test_case: The test case that served as input to the LLM program
            execution that produced output.
        quality_measure: The quality measure being evaluated.
        model: The model to use in the LLM powered quality measure.
        prompt: The prompt for the LLM powered quality measure.

    Returns:
        The result of evaluating the quality measure.
    """
    client = openai.OpenAI()
    accepted_outputs = []
    evaluation_type_prompt = ""
    if quality_measure.evaluation_type == "BINARY":
        evaluation_type_prompt = (
            "Output must be *ONLY* one of Y (for YES) or N (for NO).")
        accepted_outputs = ["Y", "N"]
    elif quality_measure.evaluation_type == "RATING_INT":
        evaluation_type_prompt = (
            "Output must be *ONLY* an integer from 1 through 5.")
        accepted_outputs = ["1", "2", "3", "4", "5"]

    encoding = tiktoken.encoding_for_model(model)
    max_tokens = 1
    logit_bias = {}
    for accepted in accepted_outputs:
        tokens = encoding.encode(accepted)
        max_tokens = max(max_tokens, len(tokens))
        logit_bias.update({token: 100 for token in tokens})

    execution_results = {
        "output": output, "inputs": inputs, "test_case": test_case}
    chat_completion = client.chat.completions.create(
        messages=[
            {"role": "system", "content": prompt.format_map(execution_results)},
            {"role": "system", "content": evaluation_type_prompt}
        ],
        model=model,
        max_tokens=max_tokens,
        logit_bias=logit_bias)
    return chat_completion.choices[0].message.content


def _execute_quality_measure_function(
    llm_program_output: Any,
    quality_measure: data_model.QualityMeasure,
    test_case: data_model.TestCase,
) -> Union[str, bool, int]:
    """Execute a function powered quality measure on the LLM program output.

    The function signature of the quality measure determines which arguments
    are passed to the function. The list of arguments is truncated to match
    the length of the function signature. Potential arguments in the order
    they are passed to the function are:
        - Output of the LLM program.
        - Test case inputs.
        - Test case itself.
        - quality_measure
    
    Args:
        llm_program_output: Output of the LLM program.
        quality_measure: Quality measure to be executed.
        test_case: Test case on which the LLM program was executed to
            produce llm_program_output.

    Returns:
        Output of the function specified in quality_measure's spec field.
    """
    callable_object = data_model.LazyCallable(quality_measure.spec)
    args = [
        llm_program_output,
        Inputs(test_case.inputs),
        test_case,
        quality_measure,
    ]
    inputs_signature = callable_object.inputs_signature
    quality_measure_output = callable_object(*args[:len(inputs_signature)])
    return quality_measure_output


def _execute_quality_measure_llm(
    llm_program_output: Any,
    quality_measure: data_model.QualityMeasure,
    test_case: data_model.TestCase,
) -> Union[str, bool, int]:
    """Execute a LLM powered quality measure on the LLM program output.

    Args:
        llm_program_output: Output of the LLM program.
        quality_measure: Quality measure.  Must have evaluator field equal
            to "LLM".
        test_case: Test case on which the LLM program was executed to
            produce llm_program_output.
    
    Raises:
        ValueError: If quality measure evaluator is not "LLM".

    Returns:
        Output of the quality measure.
    """
    if quality_measure.evaluator != "LLM":
        raise ValueError("Quality measure evaluator must be 'LLM'.")

    evaluation_type = quality_measure.evaluation_type
    if isinstance(quality_measure.spec, str):
        quality_measure_output = _execute_quality_measure_function(
            llm_program_output, quality_measure, test_case)

    elif isinstance(quality_measure.spec, dict):
        model = quality_measure.spec.get("model", "gpt-3.5-turbo")
        prompt = quality_measure.spec.get("prompt")
        if prompt is None:
            raise ValueError(
                "Quality measure spec must contain a 'prompt' key.")
        quality_measure_output = _openai_llm_quality_measure(
            llm_program_output,
            model=model,
            prompt=prompt,
            inputs=Inputs(test_case.inputs),
            quality_measure=quality_measure,
            test_case=test_case)

    if isinstance(quality_measure_output, str):
        if evaluation_type == "BINARY":
            if quality_measure_output.strip().upper() in ("YES", "Y", "TRUE"):
                quality_measure_output = True
            elif quality_measure_output.strip().upper() in ("NO", "N", "FALSE"):
                quality_measure_output = False

        elif evaluation_type == "RATING_INT":
            quality_measure_output = int(quality_measure_output)

    return quality_measure_output


class _ExecuteQualityMeasuresReturn(NamedTuple):
    """Return type for execute_quality_measures.

    See execute_quality_measures's docstring (below) for
    descriptions of fields.
    """
    direct_evaluations: List[wire_model.DirectEvaluation]
    invalid_quality_measures: List[Dict[str, Union[
        int, str, wire_model.QualityMeasureExecutionDetails]]]
    stdout_stderr: str


def execute_quality_measures(
    test_case: data_model.TestCase,
    llm_program_output: Any,
    quality_measures_with_metadata: List[
        data_model.QualityMeasureWithMetadata],
) -> _ExecuteQualityMeasuresReturn:
    """Execute quality measures on the LLM program output.

    Execute each executable quality measure in
    `quality_measures_with_metadata`. If the result is valid, add it to a
    list of direct evaluations. If the result is invalid, add it to a list
    of invalid quality measures. A quality measure's result is invalid if the
    quality measure raised an error during execution, or if the result does
    not match the quality measure's evaluation type.
 
    Args:
        test_case: Test case on which the LLM program was executed to
            produce llm_program_output.
        llm_program_output: Output of the LLM program.
        quality_measures_with_metadata: List of quality measures to be executed
            along with their IDs.
    
    Returns:
        Named tuple with the following fields:
            direct_evaluations: List of `wire_model.DirectEvaluation` objects.
            invalid_quality_measures: List of invalid quality measures. Each
                invalid quality measure is a Dictionary with the following keys:
                    "id": ID of quality measure.
                    "name": Name of quality measure.
                    "execution_details":
                        wire_model.QualityMeasureExecutionDetails object.
            stdout_stderr: Combined stdout and stderr of all quality measure
                executions.
    """
    direct_evaluations = []
    invalid_quality_measures = []
    stdout_stderr = []

    for quality_measure_with_metadata in quality_measures_with_metadata:
        # Skip quality measures that are not executable.
        if quality_measure_with_metadata.evaluator not in ("FUNCTION", "LLM"):
            continue

        # Run the (executable) quality measure.
        # TODO: https://github.com/inductor-hq/saas/issues/404 - Record
        # quality measure errors, stdout, and stderr on the backend.
        quality_measure_stdout = None
        quality_measure_stderr = None
        quality_measure_error = None
        quality_measure_output = None
        with execution.capture_stdout_stderr(
            suppress=True) as captured_streams:
            try:
                if quality_measure_with_metadata.evaluator == "FUNCTION":
                    quality_measure_output = (
                        _execute_quality_measure_function(
                            llm_program_output,
                            quality_measure_with_metadata,
                            test_case))

                elif quality_measure_with_metadata.evaluator == "LLM":
                    quality_measure_output = (
                        _execute_quality_measure_llm(
                            llm_program_output,
                            quality_measure_with_metadata,
                            test_case))

                else:
                    raise ValueError(
                        "Quality measure evaluator must be 'FUNCTION' or "
                        "'LLM'.")

            except Exception:  # pylint: disable=broad-except
                quality_measure_error = "".join(traceback.format_exc())

            quality_measure_stdout = captured_streams.stdout.getvalue()
            quality_measure_stderr = captured_streams.stderr.getvalue()
            stdout_stderr.append(captured_streams.suppressed.getvalue())

        quality_measure_execution_details = (
            wire_model.QualityMeasureExecutionDetails(
                input=llm_program_output,
                output=quality_measure_output,
                error=quality_measure_error,
                stdout=quality_measure_stdout,
                stderr=quality_measure_stderr,
            )
        )

        if quality_measure_error is not None:
            # The result of the quality measure is invalid if the quality
            # measure raised an error during execution.
            invalid_quality_measures.append({
                "id": quality_measure_with_metadata.id,
                "name": quality_measure_with_metadata.name,
                "execution_details": quality_measure_execution_details,
            })
            continue

        # Create direct evaluations for valid quality measure results.
        if (quality_measure_with_metadata.evaluation_type == "BINARY" and
            isinstance(quality_measure_output, bool)):
            direct_evaluations.append(
                wire_model.DirectEvaluation(
                    quality_measure_id=quality_measure_with_metadata.id,
                    value_bool=quality_measure_output))
        elif (quality_measure_with_metadata.evaluation_type == "RATING_INT" and
              isinstance(quality_measure_output, int) and
              # Required to prevent `bool` from being interpreted as
              # `int`, since `bool` is a subclass of `int`.
              not isinstance(quality_measure_output, bool)):
            direct_evaluations.append(
                wire_model.DirectEvaluation(
                    quality_measure_id=quality_measure_with_metadata.id,
                    value_int=quality_measure_output))
        else:
            # The result of the quality measure is invalid if the result
            # does not match the quality measure's evaluation type.
            expected_output_type = (
                type(True)
                if quality_measure_with_metadata.evaluation_type == "BINARY"
                else type(5))
            quality_measure_execution_details.error = (
                f"Invalid output type. Expected output type: "
                f"{expected_output_type}. Actual output type: "
                f"{type(quality_measure_output)}")
            invalid_quality_measures.append({
                "id": quality_measure_with_metadata.id,
                "name": quality_measure_with_metadata.name,
                "execution_details": quality_measure_execution_details,
            })

    return _ExecuteQualityMeasuresReturn(
        direct_evaluations,
        invalid_quality_measures,
        "".join(stdout_stderr))



# Copyright 2023 Inductor, Inc.
"""Functionality for playground execution tasks."""

import collections
from typing import Dict, List, Set, Union

import pydantic
import rich

from inductor import config
from inductor.backend_client import wire_model
from inductor.data_model import data_model
from inductor.execution import execution, test_suite_execution


def execute_llm_program(
    execution_request: wire_model.PlaygroundExecuteLlmProgramRequest,
    llm_program_id: str,
    llm_program_snapshot_id: str,
    llm_program_fqn: str,
) -> wire_model.PlaygroundExecuteLlmProgramResponse:
    """Execute a LLM program within a playground.
    
    Args:
        execution_request: PlaygroundExecuteLlmProgramRequest object.
        llm_program_id: LLM program ID.
        llm_program_snapshot_id: LLM program snapshot ID.
        llm_program_fqn: Fully qualified name of the LLM program.
    
    Returns:
        PlaygroundExecuteLlmProgramResponse object.
    """
    execution_results = []
    for execution_spec in execution_request.specs:
        test_case_inputs = execution_spec.inputs
        hparams = execution_spec.hparams

        hparam_sets: Dict[str, Set[Union[str, int, float]]
            ] = collections.defaultdict(set)
        if hparams is not None:
            for hparam_name, hparam_value in hparams.items():
                hparam_sets[hparam_name].add(hparam_value)

        if config.verbose:
            rich.print(
                f"Executing LLM program: "
                f"{{'inputs': {test_case_inputs}, 'hparams': {hparams}}}")

        # Execute the LLM program with the given spec.
        test_case = data_model.TestCase(test_case_inputs)
        with execution.capture_default_hparams() as default_hparams:
            # Note: we use execute_test_case here as a convenience,
            # given that it provides the functionality that we require
            # for executing an LLM program (and despite the fact
            # that we are not in fact executing a test case here).
            execution_details, _, _ = (
                test_suite_execution.execute_test_case(
                    llm_program_fully_qualified_name=llm_program_fqn,
                    test_case=test_case,
                    hparams=hparams,
                    mode="PLAYGROUND",
                    suppress_stdout_stderr=False))
            # Add default_hparams to the hparam_sets.
            for hparam_name, hparam_values in default_hparams.items():
                hparam_sets[hparam_name].update(hparam_values)

        # Convert hparam_sets to hparam_specs.
        hparam_specs: List[wire_model.HparamSpec] = []
        for hparam_name, hparam_values in hparam_sets.items():
            # Only a single value for a given hparam needs to be
            # checked to determine the type, since if other values are
            # of a different type, a ValidationError will be raised.
            if isinstance(next(iter(hparam_values)), str):
                hparam_type = "TEXT"
            else:
                hparam_type = "NUMBER"

            try:
                hparam_specs.append(
                    wire_model.HparamSpec(
                        name=hparam_name,
                        type=hparam_type,
                        values=list(hparam_values)))
            except pydantic.ValidationError as error:
                rich.print(
                    f"[yellow bold][WARNING][/yellow bold] Hparam spec, "
                    f"[cyan italic]{hparam_name}[/cyan italic], "
                    f"has invalid values and its spec will not be "
                    f"updated in the playground UI. {error}")

        execution_results.append(
            wire_model.ExecutionDetailsWithPlaygroundMetadata(
                **execution_details.model_dump(),
                fingerprint=execution_spec.fingerprint,
                hparam_specs=hparam_specs))

    return wire_model.PlaygroundExecuteLlmProgramResponse(
        llm_program_id=llm_program_id,
        llm_program_snapshot_id=llm_program_snapshot_id,
        execution_details=execution_results)


def execute_test_suite(
    execution_request: wire_model.PlaygroundExecuteTestSuiteRequest,
    llm_program_fqn: str,
    auth_access_token: str,
) -> wire_model.PlaygroundExecuteTestSuiteResponse:
    """Execute a test suite, as triggered via a playground.
    
    Args:
        execution_request: PlaygroundExecuteTestSuiteRequest object.
        llm_program_fqn: Fully qualified name of the LLM program.
        auth_access_token: Auth0 access token.
    
    Returns:
        PlaygroundExecuteTestSuiteResponse object.
    """
    test_suite_id = execution_request.test_suite_id

    test_suite = data_model.TestSuite(
        id_or_name=test_suite_id,
        llm_program=llm_program_fqn)
    test_suite.import_test_cases(test_suite_id=test_suite_id)

    test_suite_run_metadata = test_suite._run(auth_access_token)  # pylint: disable=protected-access

    return wire_model.PlaygroundExecuteTestSuiteResponse(
        test_suite_run_id=test_suite_run_metadata.test_suite_run_id,
        fingerprint=execution_request.fingerprint)



# Copyright 2023 Inductor, Inc.
"""Authentication session management for Inductor client."""

import dataclasses
import datetime
import json
import os
import pathlib
import time
from typing import Any, Optional, Union
from urllib import parse
import uuid

import auth0
from auth0 import rest
import requests
import typer

from inductor import config
from inductor.backend_client import backend_client, wire_model
from inductor.data_model import data_model


class AuthError(Exception):
    """An authentication error."""

    def __init__(self, message: str):
        """Initialize an authentication error.

        Args:
            message: The error message.
        """
        super().__init__(message)


class AuthSession:
    """An auth session object that stores the access token and its expiration.

    The AuthSession can be initiated from a secret key.  If a secret key is
    not set or not inside `~/.inductor/credentials` it tries to get it from the
    environment variable `INDUCTOR_SECRET_KEY`.  If the session is expired, it
    renews automatically when the access token is requested.
    """

    def __init__(self):
        """Initialize an AuthSession object."""
        self._access_token = None
        self._expiration = None
        self._initialized = False
        self._file_path = pathlib.Path(
            pathlib.Path.home(), ".inductor", "credentials"
        )

        self.init()

    @property
    def access_token(self):
        """The access token."""
        self._renew_expired()
        return self._access_token

    @property
    def initialized(self):
        """The initialization status of the session."""
        return self._initialized

    def _set_expiration(self, value: Union[str, datetime.datetime]):
        """Sets the expiration datetime of the access token.
        
        Args:
            value: The expiration datetime of the access token.  If str is
                given, it should be in ISO format.

        Raises:
            ValueError: If the value is not str or datetime.
        """
        if isinstance(value, str):
            self._expiration = datetime.datetime.fromisoformat(value)
        elif isinstance(value, datetime.datetime):
            self._expiration = value
        else:
            raise ValueError("Invalid expiration value.")

    def is_expired(self):
        """Checks if the session is expired."""
        if self._expiration is None:
            return True
        # Buffer the expiration time by 2 seconds.
        expiration = self._expiration - datetime.timedelta(seconds=2)
        return expiration < datetime.datetime.now(datetime.timezone.utc)

    def init(
        self,
        secret_key: Optional[str] = None,
        access_token: Optional[str] = None,
        expires_in: Optional[str] = None,
    ) -> bool:
        """Initiates the session.

        It can be initiated from a secret key or from a secret key, an access
        token and its expiration.  If secret key is not set, this method tries
        to get it from the environment variable `INDUCTOR_SECRET_KEY`, or
        otherwise from the credentials file `~/.inductor/credentials` if it
        exists.  It uses the secret key to get a new access token from the
        server.

        It stores the new session in a file if `INDUCTOR_SECRET_KEY` is not set
        and the credentials file `~/.inductor/credentials` does not exist.

        Args:
            secret_key: The secret key.
            access_token: The access token.
            expires_in: The number of seconds in which the access token will
                expire.
        
        Returns:
            True if the session was initiated, False otherwise.

        Raises:
            AuthError: If an error occurs during the authentication with
                the Inductor server endpoint.
        """
        if secret_key is None:
            if config.settings.inductor_secret_key is None:
                self._initialized = self._load_session()
                return self._initialized
            secret_key = config.settings.inductor_secret_key
        if access_token is None:
            headers = {"content-type": "application/x-www-form-urlencoded"}
            if config.settings.custom_request_headers is not None:
                headers.update(config.settings.custom_request_headers)
            inductor_token_url = config.settings.inductor_api_url
            if not inductor_token_url.endswith("/"):
                inductor_token_url += "/"
            inductor_token_url += "oauth/token"
            response = requests.post(
                inductor_token_url,
                data={"secret_key": secret_key},
                headers=headers,
                timeout=5.0,
            )
            if response.ok:
                parsed = _response_content(response)
                access_token = parsed.get("access_token")
                expires_in = parsed.get("expires_in")
            else:
                error = response.json()
                raise AuthError(error.get("detail"))
        expiration = datetime.datetime.now(
            datetime.timezone.utc
        ) + datetime.timedelta(seconds=expires_in)
        self._set_expiration(expiration)
        self._access_token = access_token
        self._initialized = True
        # If the secret key is not set in config.settings,
        # save the session to the file.
        if config.settings.inductor_secret_key is None:
            self._save_session(secret_key)
        return True

    def _renew_expired(self, secret_key: Optional[str] = None):
        """Renews the session if it is expired.

        Checks if the session is expired and renews it if necessary.

        Args:
            secret_key: The secret key.

        Raises:
            AuthError: If failed to renew the session.
        """
        if self.is_expired():
            if not self.init(secret_key):
                raise AuthError("Failed to renew expired session.")

    def _save_session(self, secret_key: str):
        """Saves the session to `~/.inductor/credentials`.

        If the file does not exist, creates it.  If the file exists, overwrites
        it with the new session.  If the secret key does not start with the
        secret key prefix, prepends the prefix before saving.

        Args:
            secret_key: The secret key used to encrypt the session.
        """
        if not secret_key.startswith(config.settings.secret_key_prefix):
            secret_key = config.settings.secret_key_prefix + secret_key
        # Create directory if it does not exist.
        self._file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._file_path, "w") as file:
            content = {
                "secret_key": secret_key,
                "access_token": self.access_token,
                "expiration": self._expiration.isoformat(),
            }
            json.dump(content, file, indent=4)

    def _load_session(self) -> bool:
        """Loads the session from `~/.inductor/credentials`.

        If the file does not exist, returns False.  If the file exists, loads
        the session from it.  If the session is expired and the secret key is
        present, renews the session.

        Raises:
            ValueError: If the credentials file is invalid or if the secret
                key is invalid.

        Returns:
            True if the session was loaded, False otherwise.
        """
        if not self._file_path.exists():
            return False
        with open(self._file_path, "r") as file:
            content = json.load(file)
            self._access_token = content.get("access_token")
            self._set_expiration(content.get("expiration"))
            secret_key: Optional[str] = content.get("secret_key")
            if (
                self._access_token is None
                or self._expiration is None
                or secret_key is None
            ):
                raise ValueError("Invalid credentials file.")
            if not secret_key.startswith(config.settings.secret_key_prefix):
                raise ValueError("Invalid secret key.")
            self._renew_expired(secret_key)
        return True


def _to_auth0_response(response: requests.Response) -> rest.Response:
    """Converts a response from an Auth0 endpoint to an Auth0 response.

    Args:
        response: The response from an Auth0 endpoint.

    Returns:
        An Auth0 response object.
    """
    if not response.text:
        return rest.EmptyResponse(response.status_code)
    try:
        return rest.JsonResponse(response)
    except ValueError:
        return rest.PlainResponse(response)


def _response_content(response: requests.Response) -> Any:
    """Extracts and returns response's content.

    Args:
        response: The response from an Auth0 endpoint.

    Returns:
        The content of the response.
    """
    return _to_auth0_response(response).content()


@dataclasses.dataclass
class _DeviceCodeResponse:
    """A response from the Auth0 device code endpoint."""

    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: int


@dataclasses.dataclass
class _TokenResponse:
    """A response from the Auth0 token endpoint."""

    access_token: str
    refresh_token: str
    id_token: str
    scope: str
    expires_in: int
    token_type: str


def _get_auth0_client_id():
    """Returns the auth0 client id.

    Users in self hosted vs inductor hosted have different auth0 apps/clients.
    Decide which to use based on the inductor api urls.
    """
    if (config.settings.inductor_hosted_api_url ==
        config.settings.inductor_api_url):
        return config.settings.auth0_client_id
    else:
        return config.settings.auth0_self_hosted_client_id


def get_device_code() -> _DeviceCodeResponse:
    """Gets a device code from Auth0.

    Returns:
        A device code response.
    """
    data = {
        "client_id": _get_auth0_client_id(),
        "audience": config.settings.inductor_hosted_api_url,
        "scope": "profile email openid offline_access",
    }
    headers = {"content-type": "application/x-www-form-urlencoded"}
    device_code_url = parse.urljoin(
        f"https://{config.settings.auth0_domain}", "/oauth/device/code")
    response = requests.post(
        device_code_url,
        data=data,
        headers=headers,
        timeout=5.0,
    )
    try:
        parsed = _response_content(response)
        return _DeviceCodeResponse(**parsed)
    except auth0.Auth0Error as e:
        raise AuthError(e.message) from e


def poll_tokens(
    auth0_device_id: str,
    device_code: str,
    interval: int
) -> _TokenResponse:
    """Polls Auth0 /oauth/token endpoint.

    Poll the Auth0 /oauth/token endpoint using device_code grant type until
    the tokens (refresh_token and access_token) are ready.

    Args:
        auth0_device_id: The ID to be used to identify the device to Auth0.
        device_code: The device code.
        interval: The polling interval (seconds).

    Returns:
        A token response.

    Raises:
        AuthError: If an error that is not "authorization_pending" or
            "slow_down" occurs during the polling.
    """
    headers = {
        "content-type": "application/x-www-form-urlencoded",
        # The user-agent is used to identify the device.
        "user-agent": auth0_device_id
    }
    data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code,
        "client_id": _get_auth0_client_id(),
    }
    while True:
        try:
            token_url = parse.urljoin(
                f"https://{config.settings.auth0_domain}", "/oauth/token")
            response = requests.post(
                token_url,
                data=data,
                headers=headers,
                timeout=5.0,
            )
            if response.ok:
                parsed = _response_content(response)
                return _TokenResponse(**parsed)
        except auth0.Auth0Error as e:
            if e.error_code == "authorization_pending":
                time.sleep(interval)
            elif e.error_code == "slow_down":
                time.sleep(interval)
            elif e.error_code == "expired_token":
                raise AuthError("Authentication failed: login page timed out. "  # pylint: disable=raise-missing-from
                                "Please try again.")
            else:
                raise AuthError("Authentication failed.")  # pylint: disable=raise-missing-from


def get_auth_session() -> AuthSession:
    """Returns a valid and not expired authenticated session.

    If the session is expired, the user is prompted to authenticate.  If
    the authentication fails, the program is aborted.
    
    Returns:
        An AuthSession object.
    """
    try:
        session = AuthSession()
        if not session.initialized:
            auth0_device_id = str(uuid.uuid4())
            device_code = get_device_code()
            typer.echo(f"Please go to {device_code.verification_uri_complete} "
                        "and log in, in order to securely authenticate the "
                        "Inductor CLI.")
            typer.echo("After you have logged in, please return here.")
            # If the user is running the CLI in VS Code, then we do not display
            # a prompt in the terminal to gate opening the authentication page
            # in the browser, as VS Code's interface will automatically display
            # such a prompt.  We also do not display such a prompt if the user
            # is running in Google Colab.
            if os.environ.get("TERM_PROGRAM") == "vscode":
                open_browser = True
            elif data_model.in_google_colab():
                open_browser = False
            else:
                open_browser = typer.confirm("Open in browser?")
            if open_browser:
                typer.launch(device_code.verification_uri_complete)
            typer.echo("Waiting for login ...")
            tokens = poll_tokens(
                auth0_device_id, device_code.device_code, device_code.interval)
            backend_client.create_api_key(
                wire_model.CreateApiKeyRequest(auth0_id=auth0_device_id),
                tokens.access_token)
            session.init(
                tokens.refresh_token,
                tokens.access_token,
                tokens.expires_in)
            typer.echo("You've successfully logged in and authenticated the "
                       "Inductor CLI!")
        return session
    except AuthError as e:
        typer.echo(e)
        raise typer.Abort() from e
    except ValueError as e:
        typer.echo(e)
        raise typer.Abort() from e



# Copyright 2023 Inductor, Inc.
"""Inductor client configuration."""

from typing import Any, Dict, Optional

import pydantic
import pydantic_settings


# TODO: #433 - Expand verbose mode.
verbose = False


class _Settings(pydantic_settings.BaseSettings):
    """Settings for the Inductor client.
    
    Settings are loaded from the following sources, in order of increasing
    precedence:
    1. Initialized values in this class
    2. Environment variables
    
    Attributes:
        auth0_domain: The Auth0 domain.
        auth0_client_id: The Auth0 client ID.
        auth0_self_hosted_client_id: The Auth0 client ID for the self hosted
            application.
        inductor_api_url: The Inductor API URL.
        inductor_hosted_api_url: The Inductor-hosted API URL.
        inductor_secret_key: The user's Inductor secret key.  If this is set,
            the client will use this secret key for authentication instead of
            prompting the user to authenticate.
        secret_key_prefix: The prefix for the secret key.
        inductor_logger_use_multiprocessing: Whether to use multiprocessing for
            `inductor.logger` (i.e.,
            `inductor.execution.live_execution.logger`).
        custom_request_headers: Custom request headers to include in all
            requests to the Inductor backend.
    """

    auth0_domain: str = "login.inductor.ai"
    auth0_client_id: str = "swO2PSUzuKcoYl5kJHjQVlVF6jnMWC6Z"
    auth0_self_hosted_client_id: str = "YFeXBAUixEVB3e9sM5e7XDZr4FEHH5XX"
    inductor_api_url: str = "https://app.inductor.ai"
    inductor_hosted_api_url: str = "https://app.inductor.ai"
    inductor_secret_key: Optional[str] = None
    secret_key_prefix: str = "isk_"
    inductor_logger_use_multiprocessing: bool = True
    custom_request_headers: Optional[pydantic.Json[Dict[str, Any]]] = (
        pydantic.Field(None, alias=pydantic.AliasChoices(
            "inductor_custom_request_headers", "custom_request_headers")))

    model_config = pydantic_settings.SettingsConfigDict(extra="ignore")


# Global settings object.
# Access settings values directly from this object (without locally caching
# values) to ensure that the settings are always up-to-date.
settings = _Settings()



# Copyright 2023 Inductor, Inc.
"""Functionality for live execution tasks."""

from concurrent import futures
import contextlib
import contextvars
import datetime
import functools
import inspect
import random
import traceback
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from inductor import auth_session, config
from inductor.backend_client import backend_client, wire_model
from inductor.data_model import data_model
from inductor.execution import execution, quality_measure_execution


# Stores a boolean flag indicating whether the current LLM program execution is
# the primary (top-level) execution. This flag allows nested functions that are
# decorated with the logger decorator to be collapsed into a single LLM program
# execution, without sending duplicate data to the backend.
_primary_execution = contextvars.ContextVar("active_execution", default=True)
# Executor for executing quality measures on live executions (so that the user
# does not have to wait for the quality measures to complete before receiving
# their LLM program's output). This executor is created lazily (i.e., when
# the logger decorator first evaluates quality measures) and is re-used for
# all subsequent quality measure evaluations. The executor is not shut down
# until the program exits. Note that multiprocessing for live executions (and
# thus this executor) is only used if
# `config.settings.inductor_logger_use_multiprocessing` is True.
_process_pool_executor = None


@contextlib.contextmanager
def _manage_executions():
    """Manage the state of the primary execution context variable.

    Manage the state of the primary execution context variable
    (_primary_execution). If the variable is initially True, it is set to
    False and True is yielded. If the variable is initially False, False is
    yielded. On exit, the variable is restored to its original value.

    The purpose of this context manager is to allow the logger decorator to
    determine whether it is the primary (top-level) execution. This is
    necessary because the logger decorator should only send data to the
    backend if it is the primary execution. For example, when the logger
    decorator decorates a function that is called by another function also
    decorated with the logger decorator, the logger decorator should not send
    data to the backend during the inner function call.

    Yields:
        True if the primary execution context variable was True, False
        otherwise.
    """
    primary_execution = _primary_execution.get()
    if primary_execution:
        _primary_execution.set(False)
    try:
        yield primary_execution
    finally:
        _primary_execution.set(primary_execution)


def _complete_live_execution(
    *,
    primary_execution: bool,
    inputs: Optional[Dict[str, Any]] = None,
    hparams: Optional[Dict[str, Union[str, int, float]]] = None,
    output: Optional[Any] = None,
    error_str: Optional[str] = None,
    stdout: Optional[str] = None,
    stderr: Optional[str] = None,
    started_at: datetime.datetime,
    logged_values: List[wire_model.LoggedValue],
    program_details: wire_model.ProgramDetails,
    live_deployment_config_id: Optional[int],
    quality_measures_with_metadata: List[data_model.QualityMeasureWithMetadata],
    auth_access_token: str,
    llm_program_execution: data_model.LiveExecution,
):
    """Complete a LLM program live execution.

    If the LLM program execution is the primary execution:
        - Execute any given quality measures on the LLM program output.
        - Send the execution data to the backend.
    Otherwise:
        - Call the Inductor client's `log()` function to log this execution as
            part of the current overarching primary execution.

    Args:
        primary_execution: Whether this is the primary execution.
        inputs: Input arguments to the LLM program.
        hparams: Hyperparameters set for the LLM program execution.
        output: Output of the LLM program. If the output is an
            `execution.IteratorWrapper` the values yielded by the iterator
            will be used as the output.
        error_str: Error (in string form) raised by the LLM program, if any.
        stdout: Contents of stdout.
        stderr: Contents of stderr.
        started_at: Time at which the LLM program execution started.
        logged_values: Values logged by the LLM program.
        program_details: Details of the LLM program.
        live_deployment_config_id: ID of the live deployment configuration
            that is associated with the LLM program, if any.
        quality_measures_with_metadata: Quality measures to be evaluated for
            the LLM program, if `primary_execution` is True. The quality
            measures include their IDs as metadata.
        auth_access_token: Auth0 access token to be used to authenticate
            requests to the backend.
        llm_program_execution: LiveExecution object used to interact with the
            execution after the LLM program has returned. 
    """
    ended_at = datetime.datetime.now(datetime.timezone.utc)

    if isinstance(output, execution.IteratorWrapper):
        # If the output is the result of an iterator, get the actual
        # output value and any error that occurred during the iteration.
        iterator_wrapper = output
        output = iterator_wrapper._get_completed_values()  # pylint: disable=protected-access
        assert error_str is None, (
            "An error occurred during the initial execution of the LLM "
            "program, but the output was an iterator. We do not expect an "
            "output from a function that raises an error.")
        error = iterator_wrapper._iteration_error  # pylint: disable=protected-access
        error_str = str(error) if error is not None else None

    if primary_execution:

        def log_live_execution(
            direct_evaluations: Optional[List[
                wire_model.DirectEvaluation]] = None,
            invalid_quality_measures: Optional[
                List[Dict[
                    str,
                    Union[int, str, wire_model.QualityMeasureExecutionDetails]
                ]]] = None):
            """Log the LLM program execution.
            
            Args:
                direct_evaluations: Direct evaluations of the LLM program
                    execution.
                invalid_quality_measures: Quality measures that could not be
                    evaluated.
            """
            # Add any invalid quality measure results to the LLM program
            # execution's logged values.
            if invalid_quality_measures is not None:
                for invalid_quality_measure in invalid_quality_measures:
                    logged_values.append(wire_model.LoggedValue(
                        value=
                        data_model.deepcopy_or_str(invalid_quality_measure),
                        description="Invalid quality measure",
                        after_complete=True))

            llm_program_execution_response = \
                backend_client.log_llm_program_execution(
                    wire_model.LogLlmProgramExecutionRequest(
                        program_details=program_details,
                        execution_details=wire_model.ExecutionDetails(
                            mode="DEPLOYED",
                            inputs=inputs if inputs else {},
                            hparams=hparams or None,
                            output=output,
                            error=error_str,
                            stdout=stdout,
                            stderr=stderr,
                            execution_time_secs=(
                                ended_at - started_at).total_seconds(),
                            started_at=started_at,
                            ended_at=ended_at,
                            logged_values=logged_values or None,
                            direct_evaluations=direct_evaluations,
                            live_deployment_config_id=live_deployment_config_id,
                        )
                    ),
                    auth_access_token
                )
            # Set the id for the LLM program execution object
            llm_program_execution._set_id(llm_program_execution_response.id) # pylint: disable=protected-access

        # Execute quality measures, if any.
        if (config.settings.inductor_logger_use_multiprocessing and
            quality_measures_with_metadata):
            global _process_pool_executor  # pylint: disable=global-variable-not-assigned
            # Create the process pool executor if it does not exist.
            if _process_pool_executor is None:
                # While most non-LLM quality measures are expected to
                # be fast and CPU bound, LLM quality measures are
                # expected to be slow and IO bound. Therefore, to
                # prevent potential bottlenecks during the evaluation
                # of LLM quality measures, we set the max workers
                # relatively high (10).
                _process_pool_executor = futures.ProcessPoolExecutor(
                    max_workers=10)
            future = _process_pool_executor.submit(
                quality_measure_execution.execute_quality_measures,
                test_case=data_model.TestCase(inputs),
                llm_program_output=output,
                quality_measures_with_metadata=quality_measures_with_metadata
            )
            future.add_done_callback(
                lambda future: log_live_execution(*future.result()[:2]))
        else:
            direct_evaluations, invalid_quality_measures, _ = (
                quality_measure_execution.execute_quality_measures(
                    test_case=data_model.TestCase(inputs),
                    llm_program_output=output,
                    quality_measures_with_metadata=
                        quality_measures_with_metadata
                ))
            log_live_execution(direct_evaluations, invalid_quality_measures)

    else:
        execution.log(
            {
                "llm_program": program_details.fully_qualified_name,
                "inputs": inputs if inputs else {},
                "output": output
            },
            name="Nested LLM program execution")


def logger(
    # NOTE: The function is optional to allow for the decorator to be used
    # with or without parentheses.
    original_function: Optional[Callable] = None,
    *,
    quality_measures: Optional[Union[
        data_model.QualityMeasure,
        List[data_model.QualityMeasure],
        data_model.YAMLFilePath]] = None,
    hparam_specs: Optional[Union[
        data_model.HparamSpec,
        List[data_model.HparamSpec],
        data_model.YAMLFilePath]] = None,
    return_execution: bool = False,
) -> Callable:
    """Log the inputs, outputs, and inductor.log calls of the given function.

    Use `logger` as a decorator to automatically log the arguments and return
    value of, as well as calls to inductor.log within, the decorated function.

    For example:
        @inductor.logger
        def hello_world(name: str) -> str:
            inductor.log(len(name), description="name length")
            return f"Hello {name}!"

    Args:
        original_function: Function to wrap. This argument should not
            be explicitly set when using @decorator syntax.
        quality_measures: Quality measures to be used to evaluate the outputs
            of the wrapped function. Quality measures can be passed in
            the following ways:
            - As an individual quality measure or a list of quality measures.
            - As a path to a YAML file specifying quality measures in the same
                format as used in Inductor test suite YAML files. (A test
                suite YAML file path would meet this requirement.)
        hparam_specs: Specifications of hyperparameters to be used
            when executing the wrapped function.  If and when
            `inductor.hparam(name, default_value)` is called within the
            wrapped function, the value that is returned will be determined
            by `hparam_specs`:
            - If `hparam_specs` contains an HparamSpec having the
                given name (per each HparamSpec's `name` field), then
                a random value from the list given by the HparamSpec's
                `values` field is returned.  Set the HparamSpec's `values`
                field to a list containing a single element to lock the
                hyperparameter to a specific value, or instead use a list
                of values to A/B(/C/etc) test among different values.
            - If `hparam_specs` does not contain an HparamSpec having the
                given name (per each HparamSpec's `name` field), then
                `inductor.hparam(name, default_value)` returns `default_value`.
            Hyperparameter specifications can be passed via `hparam_specs`
            in the following ways:
            - As an individual hyperparameter specification or a list of
                hyperparameter specifications.
            - As a path to a YAML file giving hyperparameter specifications
                in the same format as used in Inductor test suite YAML
                files. (A test suite YAML file path would meet this
                requirement.)
        return_execution: If set to True, then the wrapped function will
            be modified to return a tuple containing two elements:
            (the underlying wrapped function's return value, an Inductor
            LiveExecution object representing the wrapped function's
            execution).  The LiveExecution object's `log` method can then
            be used to log arbitrary values associated with an execution
            even after the execution has completed (e.g., to log subsequent
            end-user feedback).  If return_execution is False, then the
            underlying wrapped function's return value will be returned as
            usual (i.e., without an accompanying LiveExecution instance). 

    Returns:
        Decorator.
    """
    def _logger(func: Callable) -> Callable:
        try:
            logger_auth_session = auth_session.get_auth_session()
            # Next line is present to check that an access token can be
            # successfully obtained
            logger_auth_session.access_token  # pylint: disable=pointless-statement
            llm_program = data_model.LazyCallable(func)
            program_details = llm_program.get_program_details()
        except Exception as error:  # pylint: disable=broad-except
            traceback.print_exc()
            print(
                "[ERROR] Exception occurred during setup of inductor.logger. "
                "No data will be sent to Inductor through the logger "
                f"decorator for this session. {error}")

            @functools.wraps(func)
            def return_execution_func(
                *args, **kwargs
            ) -> Tuple[Any, data_model.LiveExecution]:
                return (func(*args, **kwargs),
                        data_model.LiveExecution(None))

            if return_execution:
                return return_execution_func
            return func

        # We only attempt to create a live deployment configuration on the
        # first call to the decorated function. Creating a live deployment
        # configuration is idempotent, so it is acceptable to accidentally
        # attempt to create multiple live deployment configurations if the
        # user calls the decorated function multiple times (and a race
        # condition occurs with the `attempt_live_deployment_config_creation`
        # flag).
        # We intentionally create live deployment configurations lazily
        # (i.e., on the first call to the decorated function) instead of
        # eagerly (i.e., when the logger decorator is called) to avoid
        # creating live deployment configurations when functions are imported
        # but not called.
        attempt_live_deployment_config_creation = True
        # Once a live deployment configuration is created, we store the
        # configuration locally in `logger_config` to avoid needing to
        # attempt to create the configuration again on subsequent calls to the
        # decorated function.
        logger_config = {}

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            llm_program_execution = data_model.LiveExecution(
                logger_auth_session)

            def with_execution(
                return_obj: Any
            ) -> Union[Any,
                       Tuple[Any, data_model.LiveExecution]]:
                if return_execution:
                    return return_obj, llm_program_execution
                return return_obj

            if execution.logger_decorator_enabled:
                # Variables from the outer scope that we need to modify.
                nonlocal logger_config
                nonlocal attempt_live_deployment_config_creation

                func_result = None
                func_error = None
                func_completed = False

                try:
                    if attempt_live_deployment_config_creation:

                        quality_measure_components = []
                        if quality_measures is not None:
                            quality_measure_components = (
                                data_model.get_test_suite_components(
                                    quality_measures)["quality_measures"])
                        hparam_spec_components = []
                        if hparam_specs is not None:
                            hparam_spec_components = (
                                data_model.get_test_suite_components(
                                    hparam_specs)["hparam_specs"])

                        if (quality_measure_components or
                            hparam_spec_components):

                            # TODO: #324 - Remove duplicates from
                            # quality_measure_components and
                            # hparam_spec_components.

                            response = (
                                backend_client.create_live_deployment(
                                    wire_model.CreateLiveDeploymentRequest(
                                        program_details=program_details,
                                        quality_measures=[
                                            quality_measure.model_dump()
                                            for quality_measure
                                            in quality_measure_components
                                        ] or None,
                                        hparam_specs=[
                                            hparam_spec.model_dump()
                                            for hparam_spec
                                            in hparam_spec_components
                                        ] or None),
                                    logger_auth_session.access_token))

                            quality_measures_with_metadata = (
                                data_model.components_with_ids(
                                    quality_measure_components,
                                    response.quality_measure_ids,
                                    data_model.QualityMeasureWithMetadata
                                ))

                            logger_config.update({
                                "live_deployment_config_id":
                                    response.live_deployment_config_id,
                                "quality_measures_with_metadata":
                                    quality_measures_with_metadata,
                                "hparam_spec_components":
                                    hparam_spec_components,
                            })

                        attempt_live_deployment_config_creation = False

                    live_deployment_config_id = logger_config.get(
                        "live_deployment_config_id", None)
                    quality_measures_with_metadata = logger_config.get(
                        "quality_measures_with_metadata", [])
                    hparam_spec_components = logger_config.get(
                        "hparam_spec_components", [])

                    hparams = {
                        hs.name: random.choice(hs.values)
                        for hs in hparam_spec_components}

                    # Notes regarding the `with` clause immediately below:
                    # - We use backslashes to split the clause, rather than
                    # enclosing in parentheses, because enclosing in
                    # parentheses apparently causes a SyntaxError when running
                    # in Python 3.8.
                    # - We actually don't need to capture stdout and stderr if
                    # we are not in the primary execution. However, since
                    # stdout and stderr are not suppressed, the user will not
                    # be impacted, and we allow stdout and stdout to be
                    # captured nonetheless to simplify the code.
                    with execution.capture_logged_values() as logged_values, \
                        _manage_executions() as primary_execution, \
                        execution.capture_stdout_stderr(
                            suppress=False) as (stdout, stderr, _), \
                        execution.set_hparams(hparams, primary_execution):

                        # Get input arguments using the function's signature.
                        signature = inspect.signature(func)
                        bound_arguments = signature.bind(*args, **kwargs)
                        bound_arguments.apply_defaults()
                        processed_input_args = {}
                        for key, value in bound_arguments.arguments.items():
                            processed_input_args[key] = (
                                data_model.deepcopy_or_str(value))

                        started_at = datetime.datetime.now(
                            datetime.timezone.utc)

                        try:
                            func_result = func(*args, **kwargs)
                        except Exception as e:  # pylint: disable=broad-except
                            func_error = e
                        finally:
                            func_completed = True

                        execution_details = {
                            "primary_execution": primary_execution,
                            "inputs": processed_input_args,
                            "hparams": hparams,
                            "error_str": (
                                str(func_error)
                                if func_error is not None else None
                            ),
                            "stdout": stdout.getvalue(),
                            "stderr": stderr.getvalue(),
                            "started_at": started_at,
                            "logged_values": logged_values,
                            "program_details": program_details,
                            "live_deployment_config_id":
                                live_deployment_config_id,
                            "quality_measures_with_metadata":
                                quality_measures_with_metadata,
                            "auth_access_token": (
                                logger_auth_session.access_token),
                            "llm_program_execution": llm_program_execution,
                        }

                        if isinstance(func_result, Iterator):
                            # We cannot assume that the iterator can be
                            # re-used. Therefore we return a wrapper for the
                            # iterator that captures returned values and will
                            # complete the execution when the iterator is
                            # exhausted.
                            return with_execution(execution.IteratorWrapper(
                                func_result,
                                stop_signal_handler=_complete_live_execution,
                                stop_signal_handler_kwargs=execution_details,
                                iterator_wrapper_error_message=(
                                    "[ERROR] Exception occurred during "
                                    "iteration of the LLM program result. No "
                                    "data will be sent to Inductor as part of "
                                    "this LLM program execution. "
                                )))

                        execution_details["output"] = (
                            data_model.deepcopy_or_str(func_result))
                        _complete_live_execution(**execution_details)

                        if func_error is not None:
                            raise func_error  # pylint: disable=raise-missing-from
                        return with_execution(func_result)

                except Exception as decorator_error:  # pylint: disable=broad-except
                    traceback.print_exc()
                    print(
                        f"[ERROR] Exception occurred in inductor.logger. No "
                        f"data will be sent to Inductor as part of this LLM "
                        f"program execution. {decorator_error}")
                    if func_completed:
                        # If the LLM program execution completed, there is no
                        # need to rerun the LLM program. Rerunning the LLM
                        # program could significantly hurt performance or could
                        # cause wider issues if the LLM program has side
                        # effects.
                        if func_error is not None:
                            raise func_error  # pylint: disable=raise-missing-from
                        return with_execution(func_result)
                    return with_execution(func(*args, **kwargs))
            else:
                return with_execution(func(*args, **kwargs))
        return wrapper

    if original_function is not None:
        return _logger(original_function)

    return _logger



# Copyright 2023 Inductor, Inc.
"""Types used by backend API that powers Inductor's client (CLI and library)."""

import datetime
from typing import Any, Dict, List, Literal, Optional, TypeVar, Union

import pydantic
# Note: we import Annotated from typing_extensions, rather than from typing,
# to enable compatibility with Python 3.8.
from typing_extensions import Annotated


# Field definitons
_RESTRICTED_NAME_FIELD = pydantic.Field(
    max_length=1000, pattern=r"^[a-zA-Z0-9_-]+$")
_SUBRESOURCE_NAME_FIELD = pydantic.Field(max_length=500)


class _WireModel(pydantic.BaseModel):
    """Base class for wire models to inherit from.

    This class configures the following setting(s):
    - extra: "ignore" (i.e., ignore extra fields passed to the model)
        In the case where the backend's (latest) wire model includes new
        fields that are not present in a given client's (out-of-date) wire
        model, this setting allows the client to still communicate with the
        backend without error. A notable side effect of this setting is that
        conversion from the client's data model to the client's wire model is
        simpler. This is the default value for this setting, but it is
        explicitly set here for clarity.
    - populate_by_name: True (i.e., fields can be populated by the name
        as given by the model attribute or the alias)
    
    To ensure backwards compatibility between the backend and clients, any
    fields added to existing wire models should have default values.
    """
    model_config = pydantic.ConfigDict(extra="ignore", populate_by_name=True)


class CreateApiKeyRequest(_WireModel):
    """Request body type for create API key endpoint.

    Attributes:
        auth0_id: Auth0 ID of the API key.
    """
    auth0_id: str


class TestCase(_WireModel):
    """A test case.

    Attributes:
        inputs: Mapping from input parameter name to input value.
        output: Optionally, an example of a desired high-quality output, or
            the output that is to be considered correct, for this test case.
        description: Optionally, a description of this test case.
    """
    inputs: Dict[str, Any]
    output: Optional[Any] = None
    description: Optional[str] = None


class TestCaseWithMetadata(TestCase):
    """A test case with metadata.
    
    Attributes:
        id: ID of the test case.
    """
    id: int


class QualityMeasure(_WireModel):
    """A quality measure.
    
    Attributes:
        name: Human-readable name of this quality measure.
        evaluator: Evaluator for this quality measure.  Determines whether
            this quality measure will be evaluated by running a function,
            or via human inspection.
        evaluation_type: The type of value produced by evaluation of this
            quality measure.
        spec: Specification of the details of how to execute this quality
            measure.
            - If evaluator is "FUNCTION", then spec should give the fully
            qualified name of the function (in the format
            "my.module:my_function").
            - If evaluator is "HUMAN", then spec should give the instructions
            or question that should be displayed to human evaluators.
            - If evaluator is "LLM", then spec should either give a dictionary
            with "model" and "prompt" fields, or the fully qualified name of an
            LLM program that implements the quality measure (in the format
            "my.module:my_function").
    """
    name: str = _SUBRESOURCE_NAME_FIELD
    evaluator: Literal["FUNCTION", "HUMAN", "LLM"]
    evaluation_type: Literal["BINARY", "RATING_INT"]
    spec: Union[str, Dict[str, str]]


class QualityMeasureWithMetadata(QualityMeasure):
    """A quality measure with metadata.

    Attributes:
        id: ID of quality measure.
    """
    id: int


def check_hparam_values_type(data: Any) -> Any:
    """Ensure that given hparam values are aligned with given the hparam type.

    This function is designed to be called in a "before" model validator
    for both the data_model.HparamSpec and wire_model.HparamSpec classes.

    For example:
    ```
    @pydantic.model_validator(mode="before")
    @classmethod
    def _check_values_type(cls: _T_HparamSpec, data: Any) -> Any:
        \"\"\"Ensure that the given values and type are aligned.\"\"\"
        check_hparam_values_type(data)
    ```

    This function is for a "before" validator instead of a "after" validator
    to provide an informative error message in the case where `values`
    is not any of the possible types (and therefore fails to pass pydantic's
    type validation). A consequence of this choice is that this function
    also needs to do some additional pre-validation to ensure that the data is
    in the expected format before performing the type validation.

    Args:
        data: The data to validate.

    Returns:
        The validated data. If the data is valid and there is a key
        "hparam_type", the key is replaced with "type".
    
    Raises:
        ValueError: If the data is not in the expected format or if the values
            are not aligned with the given hparam type.
    """
    if isinstance(data, dict):
        type_validations = {
            "SHORT_STRING":
                (str, "Hparam values of type SHORT_STRING must be strings."),
            "TEXT":
                (str, "Hparam values of type TEXT must be strings."),
            "NUMBER":
                ((int, float),
                    "Hparam values of type NUMBER must be ints or floats.")
        }

        # This function is used by both data_model.HparamSpec and
        # wire_model.HparamSpec. In both cases, the keys "type" and
        # "hparam_type" are interchangeable. Thus, unifying the keys
        # before validation is safe to do here and simplifies the
        # validation logic.
        if "type" in data and "hparam_type" in data:
            raise ValueError(
                "Only one of 'type' and 'hparam_type' should be specified "
                "as they refer to the same field.")
        if "hparam_type" in data:
            data["type"] = data.pop("hparam_type")

        if "type" not in data:
            raise ValueError("Hparam type must be specified.")
        if data["type"] not in type_validations:
            raise ValueError(
                f"Invalid hparam type: {data['type']}. Must be one of "
                f"{list(type_validations.keys())}.")
        if "values" not in data:
            raise ValueError("Hparam values must be specified.")

        expected_type, error_message = type_validations.get(
            data["type"], (None, None))
        if expected_type is not None:
            if not all(
                isinstance(value, expected_type)
                for value in data["values"]):
                raise ValueError(error_message)

    return data


# Type variable for the HparamSpec class.
_T_HparamSpec = TypeVar("_T_HparamSpec", bound="HparamSpec")  # pylint: disable=invalid-name


class HparamSpec(_WireModel):
    """Specification of set of hyperparameter values to use for test suite run.

    Attributes:
        hparam_name: Name of hyperparameter.
        hparam_type: Type of hyperparameter.
        values: List of hyperparameter values.
    """
    # Aliases are used to convert from the HparamSpec data model.
    hparam_name: str = pydantic.Field(max_length=500, alias="name")
    hparam_type: Literal["SHORT_STRING", "TEXT", "NUMBER"] = pydantic.Field(
        alias="type")
    values: Union[List[str], List[int], List[float]]

    @pydantic.model_validator(mode="before")
    @classmethod
    def _check_values_type(cls: _T_HparamSpec, data: Any) -> Any:
        """Ensure that the given values and type are aligned."""
        return check_hparam_values_type(data)


class ProgramDetails(_WireModel):
    """Details of an LLM program.

    Attributes:
        fully_qualified_name: Fully qualified name of the LLM program.
        inputs_signature:  Map between input parameter names to strings
            indicating the corresponding parameter types (or to null for
            input parameters that do not have type annotations).
        program_type: Type of LLM program.
    """
    fully_qualified_name: str
    inputs_signature: Dict[str, Optional[str]]
    program_type: Literal["FUNCTION", "LANGCHAIN"]


class LoggedValue(_WireModel):
    """A logged value associated with an LLM program execution.

    Attributes:
        value: The logged value.
        description: Description of the logged value (if any).
        after_complete: Whether the logged value was logged after the LLM
            program completed (as opposed to during its execution).
    """
    value: Any
    description: Optional[str] = None
    after_complete: bool


class DirectEvaluation(_WireModel):
    """A direct evaluation of a quality measure.
    
    Attributes:
        quality_measure_id: ID of the quality measure that was evaluated.
        value_bool: The output of the quality measure, if boolean.
        value_int: The output of the quality measure, if an integer.
    """
    quality_measure_id: int
    value_bool: Optional[bool] = None
    value_int: Optional[int] = None


# Type variable for the TestSuiteImportSpec class.
_T_TestSuiteImportSpec = TypeVar(  # pylint: disable=invalid-name
    "_T_TestSuiteImportSpec", bound="TestSuiteImportSpec")


class TestSuiteImportSpec(_WireModel):
    """Specification for importing test suite components.
    
    Exactly one of test_suite_id or meta_test_suite should be non-None.
    
    Attributes:
        test_suite_id: ID of a test suite from which to import components.
        meta_test_suite: Meta test suite specification in the format
            "<test_suite_run_id>:<quality_measure_id>".
    """
    test_suite_id: Optional[int] = None
    meta_test_suite: Optional[str] = None

    @pydantic.model_validator(mode="after")
    def check_test_suite_id_xor_meta_test_suite(self) -> "TestSuiteImportSpec":
        """Ensure that either test_suite_id or meta_test_suite is specified.
        
        Returns:
            self
        """
        if (self.test_suite_id is None) == (self.meta_test_suite is None):
            raise ValueError(
                "Exactly one of test_suite_id or meta_test_suite must be "
                "specified.")
        return self

    @pydantic.field_validator("meta_test_suite", mode="after")
    @classmethod
    def _validate_meta_test_suite(
        cls: _T_TestSuiteImportSpec, v: Optional[str]) -> Optional[str]:
        """Validate the meta_test_suite field.
        
        Args:
            cls: The TestSuiteImportSpec class.
            v: The meta_test_suite value to be validated.
            
        Raises:
            ValueError if v does not adhere to the required format (per
            attribute description in class docstring above).
            
        Returns:
            v
        """
        if v is not None:
            try:
                if len(list(map(int, v.split(":")))) != 2:
                    raise ValueError
            except (TypeError, ValueError) as error:
                raise ValueError(
                    "meta_test_suite must be in the format "
                    "<test_suite_run_id>:<quality_measure_id>.") from error
        return v


# TODO: https://github.com/inductor-hq/saas/issues/404 - This class is
# currently not used to transmit data to the backend, but instead is a
# placeholder until the backend supports recording quality measure
# execution details (i.e. errors, stdout, and stderr).
class QualityMeasureExecutionDetails(_WireModel):
    """Details of a quality measure execution.

    Attributes:
        input: The input to the quality measure.
        output: The output of the quality measure, if available.  Should be
            None if the execution terminated with an error.
        error: If an error occurred during quality measure execution, this
            field should contain a specification of the error.
        stdout: The content printed to stdout during the quality measure's
            execution.
        stderr: The content printed to stderr during the quality measure's
            execution.
    """
    input: Any
    output: Optional[Any] = None
    error: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class ExecutionDetails(_WireModel):
    """An LLM program execution.

    Attributes:
        mode: Mode in which this execution was performed (e.g., via CLI, or as
            part of a deployment), if known.
        inputs: JSON-serializable mapping from input argument name to input
            value.
        hparams: JSON-serializable mapping from hyperparameter name to
            hyperparameter value.
        output: The output of the LLM program, if available.  Should be None
            if the execution terminated with an error, without producing an
            output.
        error: If an error occurred during LLM program execution, this column
            should contain a specification of the error.
        stdout: The content printed to stdout during the LLM program's
            execution.
        stderr: The content printed to stderr during the LLM program's
            execution.
        execution_time_secs: The total wall clock time elapsed during this
            execution.
        started_at: The timestamp at which the execution started.
        ended_at: The timestamp at which the execution ended (whether
            successfully or due to an error).
        logged_values: The values logged during the LLM program's execution.
        direct_evaluations: The direct evaluations performed during the LLM
            program's execution.
        live_deployment_config_id: The ID of the live deployment config
            associated with this execution, if any.
    """
    mode: Literal["CLI", "DEPLOYED", "PLAYGROUND"]
    inputs: Dict[str, Any]
    hparams: Optional[Dict[str, Union[str, int, float]]] = None
    output: Optional[Any] = None
    error: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    execution_time_secs: float
    started_at: datetime.datetime
    ended_at: datetime.datetime
    logged_values: Optional[List[LoggedValue]] = None
    direct_evaluations: Optional[List[DirectEvaluation]] = None
    live_deployment_config_id: Optional[int] = None


class CreateTestSuiteRequest(_WireModel):
    """Request body for create test suite endpoint.

    Attributes:
        name: Name of test suite. Test suite names must be unique per user.
        description: Description of test suite.
    """
    name: str = _RESTRICTED_NAME_FIELD
    description: Optional[str] = None


class CreateTestSuiteResponse(_WireModel):
    """Response for create test suite endpoint upon successful creation.
    
    Attributes:
        id: ID of the created test suite.
    """
    id: int


# Type variable for the CreateTestSuiteRunRequest class.
_T_CreateTestSuiteRunRequest = TypeVar(  # pylint: disable=invalid-name
    "_T_CreateTestSuiteRunRequest", bound="CreateTestSuiteRunRequest")


class CreateTestSuiteRunRequest(_WireModel):
    """Request body for create test suite run endpoint.

    Attributes:
        test_suite_id: ID of test suite to run.
        test_suite_name: Name of test suite to run.
        test_cases: List of test cases.
        quality_measures: List of quality measures (if any).
        hparam_specs: List of hyperparameter specifications (if any).
        llm_program_details: Details of LLM program to run.
        replicas: Number of times that LLM program will be run on each pair of
            (test case, set of hyperparameters).
        parallelize: Degree of parallelism used for this run.
        started_at: The timestamp at which the test suite run started.
        test_suite_id_or_name: ID or name of test suite to run. This field is
            used for backwards compatibility with previous versions of the
            Inductor client. Use `test_suite_id` and `test_suite_name` instead.
        test_suite_import_specs: List of test suite import specifications.
        num_executions: The total number of executions that will be executed
            via running this test suite.
    """
    test_suite_id: Optional[int] = None
    test_suite_name: Optional[str] = None
    test_cases: List[TestCase]
    quality_measures: Optional[List[QualityMeasure]] = None
    hparam_specs: Optional[List[HparamSpec]] = None
    llm_program_details: ProgramDetails
    replicas: int = 1
    parallelize: int = 1
    started_at: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    test_suite_id_or_name: Optional[Union[int, str]] = None
    test_suite_import_specs: Optional[List[TestSuiteImportSpec]] = None
    num_executions: int = 0

    @pydantic.model_validator(mode="before")
    @classmethod
    def unpack_test_suite_id_or_name(
        cls: _T_CreateTestSuiteRunRequest, data: Any) -> Any:
        """Unpack test_suite_id_or_name into id or name fields."""
        if isinstance(data, dict):
            test_suite_id_or_name = data.get("test_suite_id_or_name")
            if test_suite_id_or_name is not None:
                if (isinstance(test_suite_id_or_name, int) and
                    "test_suite_id" not in data):
                    data["test_suite_id"] = test_suite_id_or_name
                elif (isinstance(test_suite_id_or_name, str) and
                      "test_suite_name" not in data):
                    data["test_suite_name"] = test_suite_id_or_name
        return data

    @pydantic.model_validator(mode="after")
    def check_test_suite_id_or_name(self):
        """Ensure that id or name is specified."""
        if self.test_suite_id is None and self.test_suite_name is None:
            raise ValueError(
                "Either test_suite_id or test_suite_name must be specified.")
        return self


class CreateTestSuiteRunResponse(_WireModel):
    """Response for create test suite run endpoint upon successful creation.
    
    Attributes:
        test_suite_run_id: ID of the created test suite run.
        test_case_ids: IDs of the created test cases. The order of test case
            IDs is guaranteed to be the same as the order of test cases in the
            request. If test cases were imported from other test suites, this
            field will be extended with the IDs of the imported test cases in
            the same order as the `imported_test_cases` field below.
        quality_measure_ids: IDs of the quality measures included in the
            request. The order of quality measure IDs is guaranteed to be
            the same as the order of quality measures in the request.
        hparam_spec_ids: IDs of the hyperparameter specifications included in
            the request. The order of hyperparameter specification IDs is
            guaranteed to be the same as the order of hyperparameter
            specifications in the request.
        url: URL at which the created test suite run can be accessed.
        imported_test_cases: List of test cases imported based on test suite
            import specifications specified in the request.
    """
    test_suite_run_id: int
    test_case_ids: List[int]
    quality_measure_ids: List[int]
    hparam_spec_ids: List[int]
    url: str
    imported_test_cases: Optional[List[TestCase]] = None


class LogTestCaseExecutionRequest(_WireModel):
    """Request body for log test case execution endpoint.

    Attributes:
        test_suite_run_id: ID of test suite run as part of which this
            execution occurred (if any).
        test_case_id: ID of test case on which this execution occurred.
        test_case_replica_index: Index of execution replica on test case given
            by test_case_id. Replica indices should start at zero.
        execution_details: Details of the test case execution.
    """
    test_suite_run_id: int
    test_case_id: int
    test_case_replica_index: int
    execution_details: ExecutionDetails


class CompleteTestSuiteRunRequest(_WireModel):
    """Request body for complete test suite run endpoint.

    Contains all remaining fields required by TestSuiteRun backend data model
    that have not already been logged.

    Attributes:
        test_suite_run_id: ID of the test suite run to complete.
        ended_at: The timestamp at which the test suite run ended.
    """
    test_suite_run_id: int
    ended_at: datetime.datetime


class LogLlmProgramExecutionRequest(_WireModel):
    """Request body for log LLM program execution endpoint.

    Attributes:
        program_details: Details of the LLM program.
        execution_details: Details of the LLM program execution.
    """
    program_details: ProgramDetails
    execution_details: ExecutionDetails


class LogLlmProgramExecutionResponse(_WireModel):
    """Response for log LLM program execution upon successful creation.

    Attributes:
        id: ID of the created LLM program execution.
    """
    id: int


class CreateLiveDeploymentRequest(_WireModel):
    """Request body for create live deployment endpoint.

    Attributes:
        program_details: Details of the LLM program to be
            associated with the live deployment.
        quality_measures: List of quality measures (if any).
        hparam_specs: List of hyperparameter specifications (if any).
    """
    program_details: ProgramDetails
    quality_measures: Optional[List[QualityMeasure]] = None
    hparam_specs: Optional[List[HparamSpec]] = None


class CreateLiveDeploymentResponse(_WireModel):
    """Response for create live deployment endpoint.

    Attributes:
        live_deployment_config_id: ID of the created live deployment
            configuration, if any.
        quality_measure_ids: IDs of the quality measures included in the
            request. The order of quality measure IDs is guaranteed to be
            the same as the order of quality measures in the request.
        hparam_spec_ids: IDs of the hyperparameter specifications included in
            the request. The order of hyperparameter specification IDs is
            guaranteed to be the same as the order of hyperparameter
            specifications in the request.
    """
    live_deployment_config_id: Optional[int]
    quality_measure_ids: Optional[List[int]]
    hparam_spec_ids: Optional[List[int]]


class GetTestSuiteRunQualityMeasuresRequest(_WireModel):
    """Request body for get test suite run quality measures endpoint.

    Attributes:
        quality_measure_ids: IDs of quality measures.
    """
    quality_measure_ids: List[int]


class GetTestSuiteRunQualityMeasuresResponse(_WireModel):
    """Response for get test suite run quality measures endpoint.

    Attributes:
        quality_measures: Quality measures (with metadata) corresponding to
            quality_measure_ids specified in request body.
    """
    quality_measures: List[QualityMeasureWithMetadata]


class LogValuesForExecutionRequest(_WireModel):
    """Request body for logging values post execution.
    
    Attributes:
        execution_id: ID mapping to an execution of an LLM program.
        logged_values: List of LoggedValue objects to be logged for
            the execution given by execution_id.
    """
    execution_id: int
    logged_values: List[LoggedValue]


class GetPlaygroundRequest(_WireModel):
    """Request body for get playground endpoint.

    Attributes:
        llm_program_details: Details of the LLM program to
            associate with the playground.
    """
    llm_program_details: ProgramDetails


class GetPlaygroundResponse(_WireModel):
    """Response for get playground endpoint.

    Attributes:
        playground_id: ID of the playground associated with the user ID and
            LLM program given by the request.
        llm_program_id: ID of the LLM program associated with the playground.
        llm_program_snapshot_id: ID of the LLM program snapshot associated
            with the playground.
    """
    playground_id: int
    llm_program_id: int
    llm_program_snapshot_id: int


class PlaygroundExecuteLlmProgramSpec(_WireModel):
    """Specification for executing an LLM program via a playground.

    Attributes:
        inputs: JSON-serializable mapping from input argument name to input
            value.
        hparams: JSON-serializable mapping from hyperparameter name to
            hyperparameter value.
        fingerprint: Unique identifier for this spec, if any.
    """
    inputs: Dict[str, Any]
    hparams: Optional[Dict[str, Union[str, int, float]]] = None
    fingerprint: Optional[str] = None


class PlaygroundExecuteLlmProgramRequest(_WireModel):
    """Request body for playground execute LLM program endpoint.

    Attributes:
        message_discriminator: Discriminator for this playground request,
            for use in distinguishing object types after JSON serialization.
        specs: List of LLM program execution specifications.
    """
    message_discriminator: Literal["PlaygroundExecuteLlmProgramRequest"] = (
        "PlaygroundExecuteLlmProgramRequest")
    specs: List[PlaygroundExecuteLlmProgramSpec]


class PlaygroundExecuteTestSuiteRequest(_WireModel):
    """Request body for playground execute test suite endpoint.

    Attributes:
        message_discriminator: Discriminator for this playground request,
            for use in distinguishing object types after JSON serialization.
        test_suite_id: ID of the test suite to execute.
        fingerprint: Unique identifier for this execution, if any.
    """
    message_discriminator: Literal["PlaygroundExecuteTestSuiteRequest"] = (
        "PlaygroundExecuteTestSuiteRequest")
    test_suite_id: int
    fingerprint: Optional[str] = None


# Type adapter to distinguish between playground request types.
PlaygroundRequestAdapter = pydantic.TypeAdapter(Annotated[Union[
    PlaygroundExecuteLlmProgramRequest, PlaygroundExecuteTestSuiteRequest],
    pydantic.Field(discriminator="message_discriminator")])


class ExecutionDetailsWithPlaygroundMetadata(ExecutionDetails):
    """An LLM program execution with playground metadata.

    Attributes:
        fingerprint: Unique identifier for this execution, if any.
        hparam_specs: List of hyperparameter specifications associated with
            this execution. The hparam specs are constructed from hparams in
            the execution spec and any default hparam values that were used in
            the execution.
    """
    fingerprint: Optional[str]
    hparam_specs: Optional[List[HparamSpec]] = None


class PlaygroundExecuteLlmProgramResponse(_WireModel):
    """Response for playground execute LLM program endpoint.

    Attributes:
        message_discriminator: Discriminator for this playground response,
            for use in distinguishing object types after JSON serialization.
        llm_program_id: ID of the LLM program associated with the executions.
        llm_program_snapshot_id: ID of the LLM program snapshot associated
            with the executions.
        execution_details: List of LLM program executions with fingerprint.
            The executions in this list should correspond to the specs in the
            request that triggered this response.
    """
    message_discriminator: Literal["PlaygroundExecuteLlmProgramResponse"] = (
        "PlaygroundExecuteLlmProgramResponse")
    llm_program_id: int
    llm_program_snapshot_id: int
    execution_details: List[ExecutionDetailsWithPlaygroundMetadata]


class PlaygroundExecuteTestSuiteResponse(_WireModel):
    """Response for playground execute test suite endpoint.

    Attributes:
        message_discriminator: Discriminator for this playground response,
            for use in distinguishing object types after JSON serialization.
        test_suite_run_id: ID of the test suite run that was created.
        fingerprint: Unique identifier for this execution, if any.
    """
    message_discriminator: Literal["PlaygroundExecuteTestSuiteResponse"] = (
        "PlaygroundExecuteTestSuiteResponse")
    test_suite_run_id: int
    fingerprint: Optional[str] = None


# Type adapter to distinguish between playground response types.
PlaygroundResponseAdapter = pydantic.TypeAdapter(Annotated[Union[
    PlaygroundExecuteLlmProgramResponse, PlaygroundExecuteTestSuiteResponse],
    pydantic.Field(discriminator="message_discriminator")])
